#  Live_Phase_Maps.py is based off of the example code Trigger.py.
#  Live_Phase_Maps.py records a user specified amount of images and displays
#  live phase maps.

import os

import numpy
import PySpin
import matplotlib.pyplot as plt
import time
import argparse
from PIL import Image
from numpy import array, empty, ravel, where, ones, reshape, arctan2
from matplotlib.pyplot import plot, draw, show, ion


parser = argparse.ArgumentParser(description="Initialize HPC settings")

parser.add_argument('-exp', help='The exposure time of the camera in micro seconds. Bounds: 5 - 25000.',
                    type = float, default=4000)
parser.add_argument('-gain', help='The gain of the camera in dB. Bounds: 0 - 47.',
                    type = float, default = 0)
parser.add_argument('-NI', help='The number of images to be collected by the camera.',
                    type = int, default = 400)
parser.add_argument('-PSA', help='Phase shift algorithm to use. Currently only Novak', default = 'Novak')

args = parser.parse_args()

exp = args.exp
gain = args.gain
NUM_IMAGES = args.NI
phase = args.PSA #this variable currently serves no purpuse


def configure_exp_gain(cam,exp,gain):
    #this function sets the exposure and gain as well as explicitly setting the ADC to 12 bit
    try:
        result = True
        nodemap = cam.GetNodeMap()

        # setting exposure
        if cam.ExposureAuto.GetAccessMode() != PySpin.RW:
            print('Unable to disable automatic exposure. Aborting...')
            return False

        cam.ExposureAuto.SetValue(PySpin.ExposureAuto_Off)
        #print('Automatic exposure disabled...')

        if cam.ExposureTime.GetAccessMode() != PySpin.RW:
            print('Unable to set exposure time. Aborting...')
            return False

        exposure_time_to_set = exp #measured in us
        exposure_time_to_set = min(cam.ExposureTime.GetMax(), exposure_time_to_set)
        cam.ExposureTime.SetValue(exposure_time_to_set)
        print('Shutter time set to %s us...\n' % exposure_time_to_set)


        # setting gain
        if cam.GainAuto.GetAccessMode() != PySpin.RW:
            print('Unable to disable automatic Gain. Aborting...')
            return False

        cam.GainAuto.SetValue(PySpin.GainAuto_Off)

        if cam.Gain.GetAccessMode() != PySpin.RW:
            print('Unable to set gain. Aborting...')
            return False

        gain_to_set = gain #measured in dB
        gain_to_set = min(cam.Gain.GetMax(), gain_to_set)
        cam.Gain.SetValue(gain_to_set)
        print('Gain set to %s dB...\n' % gain_to_set)

        #set ADC to 12 bit
        node_ADC = PySpin.CEnumerationPtr(nodemap.GetNode('AdcBitDepth'))
        if not PySpin.IsAvailable(node_ADC) or not PySpin.IsWritable(node_ADC):
            print('Unable to get ADC (node retrieval). Aborting...')
            return False

        node_ADC_setting = node_ADC.GetEntryByName('Bit12')
        if not PySpin.IsAvailable(node_ADC_setting) or not PySpin.IsReadable(
                node_ADC_setting):
            print('Unable to set ADC (enum entry retrieval). Aborting...')
            return False
        node_ADC.SetIntValue(node_ADC_setting.GetValue())

        #setting the pixel format
        node_pixel_format = PySpin.CEnumerationPtr(nodemap.GetNode('PixelFormat'))
        if PySpin.IsAvailable(node_pixel_format) and PySpin.IsWritable(node_pixel_format):

            # Retrieve the desired entry node from the enumeration node
            node_pixel_format_mono8 = PySpin.CEnumEntryPtr(node_pixel_format.GetEntryByName('Mono12p'))
            if PySpin.IsAvailable(node_pixel_format_mono8) and PySpin.IsReadable(node_pixel_format_mono8):

                # Retrieve the integer value from the entry node
                pixel_format_mono8 = node_pixel_format_mono8.GetValue()

                # Set integer as new value for enumeration node
                node_pixel_format.SetIntValue(pixel_format_mono8)

                print('Pixel format set to %s...' % node_pixel_format.GetCurrentEntry().GetSymbolic())

                #check bits per pixel
                ##if PySpin.IsAvailable(node_pixel_size) and PySpin.IsReadable(node_pixel_size):
                    #print(node_pixel_size)

            else:
                print('Pixel format mono 12 not available...')

        else:
            print('Pixel format not available...')


    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        result = False

    return result


class TriggerType:
    #SOFTWARE = 1
    HARDWARE = 2


CHOSEN_TRIGGER = TriggerType.HARDWARE

def configure_trigger(cam):
    """
    This function configures the camera to use a trigger. First, trigger mode is
    set to off in order to select the trigger source. Once the trigger source
    has been selected, trigger mode is then enabled, which has the camera
    capture only a single image upon the execution of the chosen trigger.

     :param cam: Camera to configure trigger for.
     :type cam: CameraPtr
     :return: True if successful, False otherwise.
     :rtype: bool
    """
    result = True

    print('*** CONFIGURING TRIGGER ***\n')


    try:
        # Ensure trigger mode off
        # The trigger must be disabled in order to configure whether the source
        # is software or hardware.
        nodemap = cam.GetNodeMap()
        node_trigger_mode = PySpin.CEnumerationPtr(nodemap.GetNode('TriggerMode'))
        if not PySpin.IsAvailable(node_trigger_mode) or not PySpin.IsReadable(node_trigger_mode):
            print('Unable to disable trigger mode (node retrieval). Aborting...')
            return False

        node_trigger_mode_off = node_trigger_mode.GetEntryByName('Off')
        if not PySpin.IsAvailable(node_trigger_mode_off) or not PySpin.IsReadable(node_trigger_mode_off):
            print('Unable to disable trigger mode (enum entry retrieval). Aborting...')
            return False

        node_trigger_mode.SetIntValue(node_trigger_mode_off.GetValue())

        print('Trigger mode disabled...')

        # Select trigger source
        # The trigger source must be set to hardware or software while trigger
        # mode is off.
        node_trigger_source = PySpin.CEnumerationPtr(nodemap.GetNode('TriggerSource'))
        if not PySpin.IsAvailable(node_trigger_source) or not PySpin.IsWritable(node_trigger_source):
            print('Unable to get trigger source (node retrieval). Aborting...')
            return False

        if CHOSEN_TRIGGER == TriggerType.HARDWARE:
            node_trigger_source_hardware = node_trigger_source.GetEntryByName('Line3') #Line 3 is the 40 Hzclock genrator signal

            if not PySpin.IsAvailable(node_trigger_source_hardware) or not PySpin.IsReadable(
                    node_trigger_source_hardware):
                print('Unable to set trigger source (enum entry retrieval). Aborting...')
                return False
            node_trigger_source.SetIntValue(node_trigger_source_hardware.GetValue())

        # Turn trigger mode on
        # Once the appropriate trigger source has been set, turn trigger mode
        # on in order to retrieve images using the trigger.
        node_trigger_mode_on = node_trigger_mode.GetEntryByName('On')
        if not PySpin.IsAvailable(node_trigger_mode_on) or not PySpin.IsReadable(node_trigger_mode_on):
            print('Unable to enable trigger mode (enum entry retrieval). Aborting...')
            return False

        node_trigger_mode.SetIntValue(node_trigger_mode_on.GetValue())
        print('Trigger mode turned back on...')

    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        return False

    return result


def grab_next_image_by_trigger(nodemap, cam):  #is this necessary if we don't have a software trigger???
    """
    This function acquires an image by executing the trigger node.

    :param cam: Camera to acquire images from.
    :param nodemap: Device nodemap.
    :type cam: CameraPtr
    :type nodemap: INodeMap
    :return: True if successful, False otherwise.
    :rtype: bool
    """
    try:
        result = True
        # Use trigger to capture image
        # The software trigger only feigns being executed by the Enter key;
        # what might not be immediately apparent is that there is not a
        # continuous stream of images being captured; in other examples that
        # acquire images, the camera captures a continuous stream of images.
        # When an image is retrieved, it is plucked from the stream.


        if CHOSEN_TRIGGER == TriggerType.HARDWARE:
            a = 2 #I do not understand the need for this but it works...

    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        return False

    return result

def Novak_phase_no_mask(listo):
    #phase of each pixel, assuming: list of five images read in, equally centered and sized
    arr1 = numpy.ravel(numpy.array(listo[0],dtype='int'))
    arr2 = numpy.ravel(numpy.array(listo[1],dtype='int')) #converts to numpy arrays for faster operation
    arr3 = numpy.ravel(numpy.array(listo[2],dtype='int'))
    arr4 = numpy.ravel(numpy.array(listo[3],dtype='int'))
    arr5 = numpy.ravel(numpy.array(listo[4],dtype='int'))
    phase = numpy.empty(388800)

    p1 = arr1
    p2 = arr2
    p3 = arr3
    p4 = arr4
    p5 = arr5

    den = 2*p3-p1-p5

    A = p2-p4

    B = p1-p5+10

    num = numpy.sqrt(abs(4*A**2-B**2))

    pm = numpy.sign(A)

    pha = numpy.arctan2(pm*num,den)

    phase = pha

    phase = numpy.reshape(phase,(540,720))

    return phase

def Novak_phase(listo):
    #phase of each pixel, assuming: list of five images read in, equally centered and sized
    arr1 = numpy.ravel(numpy.array(listo[0],dtype='int'))
    arr2 = numpy.ravel(numpy.array(listo[1],dtype='int')) #converts to numpy arrays for faster operation
    arr3 = numpy.ravel(numpy.array(listo[2],dtype='int'))
    arr4 = numpy.ravel(numpy.array(listo[3],dtype='int'))
    arr5 = numpy.ravel(numpy.array(listo[4],dtype='int'))
    phase = numpy.empty(388800)

    mask = numpy.ones(388800,dtype=bool)

    cuts = numpy.where(arr1 < 3000)

    mask[cuts] = False

    p1 = arr1[mask]
    p2 = arr2[mask]
    p3 = arr3[mask]
    p4 = arr4[mask]
    p5 = arr5[mask]

    den = 2*p3-p1-p5

    A = p2-p4

    B = p1-p5

    num = numpy.sqrt(abs(4*A**2-B**2))

    pm = numpy.sign(A)

    pha = numpy.arctan2(pm*num,den)

    phase[~mask] = 0
    phase[mask] = pha

    phase = numpy.reshape(phase,(540,720))

    return phase

def fourpointphase(listo):
    #phase of each pixel, assuming: list of four images read in, equally centered and sized
    arr1 = numpy.ravel(numpy.array(listo[0],dtype='int'))
    arr2 = numpy.ravel(numpy.array(listo[1],dtype='int')) #converts to numpy arrays for faster operation
    arr3 = numpy.ravel(numpy.array(listo[2],dtype='int'))
    arr4 = numpy.ravel(numpy.array(listo[3],dtype='int'))

    phase = numpy.empty(388800)

    mask = numpy.ones(388800,dtype=bool)

    cuts = numpy.where(arr1 < 450) #12 bit arrays have a factor of 16 for some reason

    mask[cuts] = False

    p1 = arr1[mask]
    p2 = arr2[mask]
    p3 = arr3[mask]
    p4 = arr4[mask]

    num = p4 - p2
    den = p1 - p3
    pha = numpy.arctan2(num,den)

    phase[~mask] = 0
    phase[mask] = pha

    phase = numpy.reshape(phase,(540,720))

    return phase

def carre_phase(listo):
    #phase of each pixel, assuming: list of four images read in, equally centered and sized
    arr1 = numpy.ravel(numpy.array(listo[0],dtype='int'))
    arr2 = numpy.ravel(numpy.array(listo[1],dtype='int')) #converts to numpy arrays for faster operation
    arr3 = numpy.ravel(numpy.array(listo[2],dtype='int'))
    arr4 = numpy.ravel(numpy.array(listo[3],dtype='int'))

    phase = numpy.empty(388800)

    mask = numpy.ones(388800,dtype=bool)

    cuts = numpy.where(arr1 < 450)

    mask[cuts] = False

    p1 = arr1[mask]
    p2 = arr2[mask]
    p3 = arr3[mask]
    p4 = arr4[mask]

    B = p1-p4
    A = p2-p3
    num = (A+B) * (3*A-B)
    num = numpy.sqrt(abs(num))
    pm = numpy.sign(A)
    den = p2 + p3 - p1 - p4
    pha = numpy.arctan2(pm*num,den)

    phase[~mask] = 0
    phase[mask] = pha

    phase = numpy.reshape(phase,(540,720))

    return phase

def acquire_images(cam, nodemap, nodemap_tldevice, NUM_IMAGES):
    """
    This function acquires and saves 10 images from a device.
    Please see Acquisition example for more in-depth comments on acquiring images.

    :param cam: Camera to acquire images from.
    :param nodemap: Device nodemap.
    :param nodemap_tldevice: Transport layer device nodemap.
    :type cam: CameraPtr
    :type nodemap: INodeMap
    :type nodemap_tldevice: INodeMap
    :return: True if successful, False otherwise.
    :rtype: bool
    """

    print('*** IMAGE ACQUISITION ***\n')
    try:
        result = True

        # Set acquisition mode to continuous
        # In order to access the node entries, they have to be casted to a pointer type (CEnumerationPtr here)
        node_acquisition_mode = PySpin.CEnumerationPtr(nodemap.GetNode('AcquisitionMode'))
        if not PySpin.IsAvailable(node_acquisition_mode) or not PySpin.IsWritable(node_acquisition_mode):
            print('Unable to set acquisition mode to continuous (enum retrieval). Aborting...')
            return False

        # Retrieve entry node from enumeration node
        node_acquisition_mode_continuous = node_acquisition_mode.GetEntryByName('Continuous')
        if not PySpin.IsAvailable(node_acquisition_mode_continuous) or not PySpin.IsReadable(
                node_acquisition_mode_continuous):
            print('Unable to set acquisition mode to continuous (entry retrieval). Aborting...')
            return False

        # Retrieve integer value from entry node
        acquisition_mode_continuous = node_acquisition_mode_continuous.GetValue()

        # Set integer value from entry node as new value of enumeration node
        node_acquisition_mode.SetIntValue(acquisition_mode_continuous)

        print('Acquisition mode set to continuous...')

        #  Begin acquiring images
        cam.BeginAcquisition()

        print('Acquiring images...')

        #  Retrieve device serial number for filename
        #
        #  *** NOTES ***
        #  The device serial number is retrieved in order to keep cameras from
        #  overwriting one another. Grabbing image IDs could also accomplish
        #  this.
        device_serial_number = ''
        node_device_serial_number = PySpin.CStringPtr(nodemap_tldevice.GetNode('DeviceSerialNumber'))
        if PySpin.IsAvailable(node_device_serial_number) and PySpin.IsReadable(node_device_serial_number):
            device_serial_number = node_device_serial_number.GetValue()
            print('Device serial number retrieved as %s...' % device_serial_number)



        print('')
        picList = []
        phaseList = []


        # Retrieve, convert, and save images
        for i in range(NUM_IMAGES):
            try:

                #  Retrieve the next image from the trigger
                result &= grab_next_image_by_trigger(nodemap, cam)

                #  Retrieve next received image
                image_result = cam.GetNextImage()

                #  Ensure image completion
                if image_result.IsIncomplete():
                    print('Image incomplete with image status %d ...' % image_result.GetImageStatus())

                else:

                    #  Convert image to mono 12
                    #
                    #  *** NOTES ***
                    #  Images can be converted between pixel formats by using
                    #  the appropriate enumeration value. Unlike the original
                    #  image, the converted one does not need to be released as
                    #  it does not affect the camera buffer.
                    #
                    #  When converting images, color processing algorithm is an
                    #  optional parameter.
                    """
                    TO DO:
                    Change the pixel format to 12 bit.
                    """
                    image_converted = image_result.Convert(PySpin.PixelFormat_Mono16, PySpin.HQ_LINEAR)

					#add to piclist
                    imgarray = image_converted.GetNDArray()
                    picList.append(imgarray)


					#  Release image
                    image_result.Release()
                    if i%4 == 0 and i > 0:
                        if i%8 == 0:
                            #there are other phase methods that can be run...
                            faze = Novak_phase(picList) #Novak_phase_no_mask #Novak_phase #carre_phase #fourpointphase
                            phaseList.append(faze)
                            #print(faze.dtype)
                            #print(numpy.shape(faze))
                            plt.ion()
                            plt.imshow(faze, cmap = 'jet')
                            cbar = plt.colorbar()#
                            plt.clim(vmin=-numpy.pi,vmax=numpy.pi)#
                            cbar.set_label("Phase Shift (rad)")#
                            plt.xlabel("Pixels(x)")
                            plt.ylabel("Pixels(y)")
                            #plt.xlim([300,500]) If a smaller image is wanted
                            #plt.ylim([200,400])
                            plt.pause(0.00001)
                            plt.show()
                            plt.clf()

                        del picList[0:4]

            except PySpin.SpinnakerException as ex:
                print('Error: %s' % ex)
                return False

        # End acquisition
        #
        #  *** NOTES ***
        #  Ending acquisition appropriately helps ensure that devices clean up
        #  properly and do not need to be power-cycled to maintain integrity.
        cam.EndAcquisition()

    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        return False

    return result


def reset_trigger(nodemap):
    """
    This function returns the camera to a normal state by turning off trigger mode.

    :param nodemap: Transport layer device nodemap.
    :type nodemap: INodeMap
    :returns: True if successful, False otherwise.
    :rtype: bool
    """
    try:
        result = True
        node_trigger_mode = PySpin.CEnumerationPtr(nodemap.GetNode('TriggerMode'))
        if not PySpin.IsAvailable(node_trigger_mode) or not PySpin.IsReadable(node_trigger_mode):
            print('Unable to disable trigger mode (node retrieval). Aborting...')
            return False

        node_trigger_mode_off = node_trigger_mode.GetEntryByName('Off')
        if not PySpin.IsAvailable(node_trigger_mode_off) or not PySpin.IsReadable(node_trigger_mode_off):
            print('Unable to disable trigger mode (enum entry retrieval). Aborting...')
            return False

        node_trigger_mode.SetIntValue(node_trigger_mode_off.GetValue())

        print('Trigger mode disabled...')

    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        result = False

    return result


def print_device_info(nodemap):
    """
    This function prints the device information of the camera from the transport
    layer; please see NodeMapInfo example for more in-depth comments on printing
    device information from the nodemap.

    :param nodemap: Transport layer device nodemap.
    :type nodemap: INodeMap
    :returns: True if successful, False otherwise.
    :rtype: bool
    """

    print('*** DEVICE INFORMATION ***\n')

    try:
        result = True
        node_device_information = PySpin.CCategoryPtr(nodemap.GetNode('DeviceInformation'))

        if PySpin.IsAvailable(node_device_information) and PySpin.IsReadable(node_device_information):
            features = node_device_information.GetFeatures()
            for feature in features:
                node_feature = PySpin.CValuePtr(feature)
                print('%s: %s' % (node_feature.GetName(),
                                  node_feature.ToString() if PySpin.IsReadable(node_feature) else 'Node not readable'))

        else:
            print('Device control information not available.')

    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        return False

    return result


def run_single_camera(cam):
    """
    This function acts as the body of the example; please see NodeMapInfo example
    for more in-depth comments on setting up cameras.

    :param cam: Camera to run on.
    :type cam: CameraPtr
    :return: True if successful, False otherwise.
    :rtype: bool
    """
    try:
        result = True
        err = False

        # Retrieve TL device nodemap and print device information
        nodemap_tldevice = cam.GetTLDeviceNodeMap()

        result &= print_device_info(nodemap_tldevice)

        # Initialize camera
        cam.Init()

        # Retrieve GenICam nodemap
        nodemap = cam.GetNodeMap()

        #configure settings
        if configure_exp_gain(cam,exp,gain) is False:
            return False

        # Configure trigger
        if configure_trigger(cam) is False:
            return False

        # Acquire images
        result &= acquire_images(cam, nodemap, nodemap_tldevice, NUM_IMAGES)

        # Reset trigger
        result &= reset_trigger(nodemap)

        # Deinitialize camera
        cam.DeInit()

    except PySpin.SpinnakerException as ex:
        print('Error: %s' % ex)
        result = False

    return result


def main():
    #delete this later
    exp = 1000
    gain = 0
    NI = 400
    PSA = 'Novak'
    """
    Example entry point; please see Enumeration example for more in-depth
    comments on preparing and cleaning up the system.

    :return: True if successful, False otherwise.
    :rtype: bool
    """

    # Since this application saves images in the current folder
    # we must ensure that we have permission to write to this folder.
    # If we do not have permission, fail right away.
    try:
        test_file = open('test.txt', 'w+')
    except IOError:
        print('Unable to write to current directory. Please check permissions.')
        input('Press Enter to exit...')
        return False

    test_file.close()
    os.remove(test_file.name)

    result = True

    # Retrieve singleton reference to system object
    system = PySpin.System.GetInstance()

    # Get current library version
    version = system.GetLibraryVersion()
    print('Library version: %d.%d.%d.%d' % (version.major, version.minor, version.type, version.build))

    # Retrieve list of cameras from the system
    cam_list = system.GetCameras()

    num_cameras = cam_list.GetSize()

    print('Number of cameras detected: %d' % num_cameras)

    # Finish if there are no cameras
    if num_cameras == 0:
        # Clear camera list before releasing system
        cam_list.Clear()

        # Release system instance
        system.ReleaseInstance()

        print('Not enough cameras!')
        input('Done! Press Enter to exit...')
        return False

    # Run example on each camera
    for i, cam in enumerate(cam_list):

        print('Running example for camera %d...' % i)

        result &= run_single_camera(cam)
        print('Camera %d example complete... \n' % i)

    # Release reference to camera
    # NOTE: Unlike the C++ examples, we cannot rely on pointer objects being automatically
    # cleaned up when going out of scope.
    # The usage of del is preferred to assigning the variable to None.
    del cam

    # Clear camera list before releasing system
    cam_list.Clear()

	# averaging method
    #avg = Avg(picList, 160) ### 160 NUMBER OF IMAGES TO USE, SHOULD BE CONFIDENT NO BEAM stepping

    # Release system instance
    system.ReleaseInstance()

    input('Done! Press Enter to exit...')
    return result


if __name__ == '__main__':
    main()



# pygator/spec/welch_analysis.py
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import welch

def welch_psd(voltage, sampling_rate, plot=True, window='hann', average=None):
    """Compute PSD using Welch's method"""
    frequencies, psd = welch(voltage, fs=sampling_rate, nperseg=len(voltage),
                             window=window, average=average)
    # asd = np.sqrt(psd)  # amplitude spectral density

    if plot:
        plt.figure()
        plt.loglog(frequencies, np.sqrt(psd))
        plt.xlabel('Frequency [Hz]')
        plt.ylabel(r'ASD $(unit/sqrt{Hz})$')
        plt.grid(True)
        plt.show()
    
    return frequencies, psd



#%%
import pygator 
import numpy as np
import matplotlib.pyplot as plt 
import sympy as sp 
import subprocess
from pathlib import Path
git_repo_path = subprocess.check_output(['git', 'rev-parse', '--show-toplevel']).strip().decode('utf-8')
base_path = Path(git_repo_path) / "random_scripts" 
from scipy.signal import welch

# The goal is to reconstruct the phase of this beam using Novak PSA
# https://www.sciencedirect.com/science/article/pii/S0030402604702260
x=y=np.linspace(-3e-3,3e-3,400)
w0=700e-6
z=1e-2
wavelength=1064e-9
zr=np.pi*w0**2/wavelength
hg00=pygator.module.HG_mode_num(x,y,n=0,m=0,q=z+1j*zr)
hg10=pygator.module.HG_mode_num(x,y,n=1,m=0,q=z+1j*zr)


extend=[x.min()/1e-3,x.max()/1e-3,y.min()/1e-3,y.max()/1e-3]
signal=hg00['U']#+0.1*hg10['U']
fig, axs = plt.subplots(1, 2, figsize=(12, 5))

# Magnitude plot
im0 = axs[0].imshow(np.abs(signal),extent=extend, cmap='jet')
axs[0].set_title('Magnitude')
axs[0].set_xlabel("x-coordinates [mm]")
axs[0].set_xlabel("y-coordinates [mm]")
plt.colorbar(im0, ax=axs[0])

# Phase plot.. let's try to reconstruct this from Novak algorithm
im1 = axs[1].imshow(np.angle(signal),extent=extend, cmap='jet')
axs[1].set_title('Phase')
axs[1].set_xlabel("x-coordinates [mm]")
axs[1].set_xlabel("y-coordinates [mm]")
plt.colorbar(im1, ax=axs[1])
plt.suptitle("Magnitude and phase of the beam to recover")
plt.tight_layout()
plt.savefig((Path(git_repo_path) / "random_scripts"/"plots"/"beam.pdf"))

plt.show()

#%%
# Let's create 5 intensity images at 0, pi/2,pi,3pi/2,2pi

phi = np.angle(signal)              # true phase 
B = np.abs(signal)**2/np.max(np.abs(signal)**2)  # normalized intensity
A = 0*np.mean(B)                    # constant phase shift.. make it zero

deltas = [0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi]
I = [A + B * np.cos(phi + d) for d in deltas]


save_path = Path(git_repo_path) / "random_scripts" / "plots"

plt.figure(figsize=(12, 8))

for idx, img in enumerate(I):
    plt.subplot(2, 3, idx+1)
    plt.imshow(img, cmap='jet', extent=extend, origin='lower')
    plt.title(f"δ = {deltas[idx]/np.pi:.1f} π")
    plt.xlabel("x [mm]")
    plt.ylabel("y [mm]")
    plt.colorbar(fraction=0.046, pad=0.04)

# Optional: leave the 6th subplot empty
plt.subplot(2,3,6)
plt.axis('off')

plt.tight_layout()
plt.savefig(save_path / "intensity_cycle_grid.pdf")
plt.show()
# %%

def novak(lis):
    den = 2*lis[2]-lis[0]-lis[4]
    A = lis[1]-lis[3]
    B = lis[0]-lis[4]
    num = np.sqrt(abs(4*A**2-B**2))
    pm = np.sign(A)
    return np.arctan2(pm*num,den)


recovered_PSA=novak(I)
plt.imshow(recovered_PSA,cmap='jet')
plt.colorbar()
# %%


# unwrap recovered phase along both axes
recovered_unwrapped = np.unwrap(recovered_PSA, axis=0)
recovered_unwrapped = np.unwrap(recovered_unwrapped, axis=1)

plt.figure(figsize=(12, 5))

#original phase
plt.subplot(1, 2, 1)
plt.imshow(phi, cmap='jet', extent=extend, origin='lower')
plt.title("Original phase")
plt.xlabel("x [mm]")
plt.ylabel("y [mm]")
plt.colorbar(fraction=0.046, pad=0.04)

#recovered Novak phase
plt.subplot(1, 2, 2)
plt.imshow(recovered_unwrapped, cmap='jet', extent=extend, origin='lower')
plt.title("Recovered Novak phase")
plt.xlabel("x [mm]")
plt.ylabel("y [mm]")
plt.colorbar(fraction=0.046, pad=0.04)

plt.tight_layout()
plt.savefig(Path(git_repo_path) / "random_scripts" / "plots" / "recovered_phase.pdf")
plt.show()


# %%
print("Original phase: max - min: ",phi.max()-phi.min())
print("Recovered phase: max - min: ",recovered_unwrapped.max()-recovered_unwrapped.min())
# %%



# my_package/__init__.py
# from pygator.module import *
# from ._version import __version__

# pygator/__init__.py
from ._version import __version__
from . import module
from . import spectran
from . import ABCD
from . import beam_profile
from . import library


# my_package/__init__.py
# from pygator.module import *
# from ._version import __version__

# pygator/__init__.py
import numpy as np

def multiply_abcd(*matrices, reverse=True, q_in=None):
    """
    Multiply an arbitrary number of ABCD matrices and optionally 
    propagate a Gaussian beam q-parameter.

    Parameters
    ----------
    *matrices : array-like
        Any number of 2x2 numpy arrays representing ABCD matrices.
        Example: multiply_abcd(M1, M2, M3)

    reverse : bool, optional (default=True)
        If True, multiply in reverse order (optical convention: 
        beam hits last matrix first).
        If False, multiply in given order.

    q_in : complex or None, optional
        Input Gaussian beam q-parameter. If provided, the output q 
        after propagation is also returned.

    Returns
    -------
    result : ndarray
        2x2 numpy array representing the combined ABCD matrix.

    q_out : complex (optional)
        Output q-parameter if q_in is given.
    """

    # Ensure all are numpy arrays
    mats = [np.array(M, dtype=float) for M in matrices]

    # Reverse order if needed
    if reverse:
        mats = mats[::-1]

    # Multiply sequentially
    result = np.eye(2)
    for M in mats:
        result = result @ M

    if q_in is not None:
        A, B, C, D = result.ravel()
        q_out = (A * q_in + B) / (C * q_in + D)
        return result, q_out

    return result


def multiply_matrices(*matrices, reverse=False, q_in=None):
    """
    Multiply an arbitrary number of square matrices.
    Works for 2x2 (ABCD, Jones), 3x3, NxN, etc.
    
    Parameters
    ----------
    *matrices : list of ndarray
        Matrices to multiply.
    reverse : bool, optional
        If True, multiply in reverse order (last applied first).
    q_in : complex, optional
        If provided, applies ABCD transformation (only valid for 2x2 matrices).
    
    Returns
    -------
    M_total : ndarray
        Product of all matrices.
    q_out : complex or None
        If q_in is given and matrices are 2x2, return output q.
    """
    
    if reverse:
        matrices = matrices[::-1]
    
    # Ensure all matrices are numpy arrays (allowing complex)
    mats = [np.array(M, dtype=complex) for M in matrices]
    
    # Infer dimension from first matrix
    n = mats[0].shape[0]
    M_total = np.eye(n, dtype=complex)
    
    for M in mats:
        M_total = M @ M_total
    
    # If q_in is provided and we are in 2x2 land, apply ABCD transform
    if q_in is not None and n == 2:
        A, B, C, D = M_total[0,0], M_total[0,1], M_total[1,0], M_total[1,1]
        q_out = (A * q_in + B) / (C * q_in + D)
        return M_total, q_out
    
    return M_total



import cv2
import numpy as np
import argparse
import matplotlib.pyplot as plt
import tkinter as tk
from tkinter import simpledialog, messagebox
from pygator.beam_profile.fit_gaussian import fit_gaussian_roi, gaussian_2d
from pygator.beam_profile.live_camera import get_camera_and_start
from pygator.module import fit_beam_profile_ODR
import csv
import PySpin

def draw_text(image, text, pos=(10, 20), color=(255,)):
    font = cv2.FONT_HERSHEY_SIMPLEX
    cv2.putText(image, text, pos, font, 0.5, color, 1)

def get_distance_input():
    root = tk.Tk()
    root.withdraw()  # Hide the main window
    user_input = simpledialog.askstring("Distance Moved (inches)", 
                                        "Enter distance moved from previous position (inches, e.g. 1.0):")
    root.destroy()
    return user_input

def ask_to_save():
    root = tk.Tk()
    root.withdraw()
    response = messagebox.askyesno("Save Data", "Save recorded data to CSV?")
    root.destroy()
    return response

def draw_grid(img, step=50, color=(100, 100, 100), thickness=1):
    """Draws a grid over the entire image."""
    h, w = img.shape[:2]

    # Vertical lines
    for x in range(0, w, step):
        cv2.line(img, (x, 0), (x, h), color, thickness)

    # Horizontal lines
    for y in range(0, h, step):
        cv2.line(img, (0, y), (w, y), color, thickness)


def beam_profile_fit(roi_size=300, downsample=2, exposure='auto', gain='auto',
                     pixel_size_um=6.9, output_file="beam_profile.csv", mode="gray",theta_val=0
                     ,scaling_factor=1):

    pixel_size_m = pixel_size_um * 1e-6
    cam, cam_list, system = get_camera_and_start(exposure, gain)
    if cam is None:
        return

    theta_val = float(theta_val)  # ensure numeric
    theta_step = 1.0  # degrees per key press
    roi_size_step = 20.0 # increase in size of one side of roi per key press
    z_list, wx_list, wy_list, wx_std_list, wy_std_list, theta_fit_list, A_list, B_list  = [], [], [], [], [], [], [], []
    wx_temp, wy_temp = [], []
    meshgrid_cache = {}
    z_position = 0.0

    print("Live beam profiling started. Press:")
    print("  [g] Fit a Gaussian to the beam manually")
    print("  [s] Increases roi size value by 20 pixels")
    print("  [d] Decreases roi size value by 20 pixels")
    print("  [p] Manually change exposure time")
    print("  [r] Record current sample (adds to buffer)")
    print("  [R] Finalize buffer (mean/std saved to dataset)")
    print("  [n] Move camera (input distance in inches)")
    print("  [[] Decreases the true coordinate rotation angle by 1 degree")
    print("  []] Increase the true coordinate rotation angle by 1 degree")
    print("  [f] Fit data and finish")
    print("  [q] Quit without fitting")

    plt.ion()
    plt.figure("Beam Width Live Plot")

    try:
        recording = False
        gaussian_fit_boolean = False
        plt.ion()
        residual_fig,residual_ax=plt.subplots()
        residual_im=None
        while True:
                # Try to grab a frame
            try:
                image_result = cam.GetNextImage()
            except Exception as e:
                print(f"Warning: Camera read failed (camera may be disconnected): {e}")

                # Save buffered data to a temporary CSV
                temp_file = output_file.replace(".csv", "_disconnected_backup.csv")
                if len(z_list) > 0:
                    with open(temp_file, "w", newline="") as f:
                        writer = csv.writer(f)
                        writer.writerow(["z [m]", "wx [m]", "wy [m]", "wx_std [m]", "wy_std [m]", "theta_fit [deg]", "A [grayscale unit]", "B [grayscale unit]"])
                        for i in range(len(z_list)):
                            writer.writerow([z_list[i], wx_list[i], wy_list[i], wx_std_list[i], wy_std_list[i], theta_fit_list[i], A_list[i], B_list[i]])
                    print(f"Partial data saved to {temp_file}")

                break  # Exit the loop gracefully

            if image_result.IsIncomplete():
                print("Incomplete image:", image_result.GetImageStatus())
                image_result.Release()
                continue

            try:
                img = image_result.GetNDArray().copy()
            except Exception as e:
                print("Error reading image:", e)
                image_result.Release()
                continue

            image_result.Release()


            # Choose mode-based display
            if mode == 'heatmap':
                display_img = cv2.applyColorMap(img, cv2.COLORMAP_JET)
                rect_color = (255, 0, 0)  # Blue rectangle
                text_color = (255, 255, 255)
                ellipse_color = (255, 255, 255)
            else:
                display_img = img
                rect_color = (200,)       # Light gray rectangle
                text_color = (255,)
                ellipse_color = (255,)

                # params = fit_gaussian_roi(img, roi_size=roi_size, downsample=downsample, meshgrid_cache=meshgrid_cache,theta_user=theta_val)
                # x0, y0 = int(params[1]), int(params[2])
                # half = roi_size // 2
                # top_left = (max(0, x0 - half), max(0, y0 - half))
                # bottom_right = (min(img.shape[1], x0 + half), min(img.shape[0], y0 + half))

                # w_x_px = params[3]
                # w_y_px = params[4]

                # w_x_m = w_x_px * pixel_size_m/scaling_factor #1.13 scaling factor by comparing the FLIR camera to the WinCam
                # w_y_m = w_y_px * pixel_size_m/scaling_factor
                # A = params[0]

                # # Use user theta for drawing
                # theta = params[-1]
                # theta_fit=params[-2] 

                # # Choose mode-based display
                # if mode == 'heatmap':
                #     display_img = cv2.applyColorMap(img, cv2.COLORMAP_JET)
                #     rect_color = (255, 0, 0)  # Blue rectangle
                #     text_color = (255, 255, 255)
                #     ellipse_color = (255, 255, 255)
                # else:
                #     display_img = img
                #     rect_color = (200,)       # Light gray rectangle
                #     text_color = (255,)
                #     ellipse_color = (255,)
                
            #     #Adding a grid
            #     draw_grid(display_img, step=100, color=(100, 100, 100))  # gray grid
            #     # Draw ROI rectangle and ellipse
            #     cv2.rectangle(display_img, top_left, bottom_right, rect_color, 1)
            #     center = (x0, y0)
            #     axes = (int(params[3] * 2), int(params[4] * 2))
            #     cv2.ellipse(display_img, center, axes, 0, 0, 360, ellipse_color, 1)

            #     # Crosshair parameters
            #     # Use ROI size for line lengths
            #     Lx = (bottom_right[0] - top_left[0]) // 2  # half width
            #     Ly = (bottom_right[1] - top_left[1]) // 2  # half height

            #     # Function to rotate points
            #     def rotate_point(x, y, theta, x0=0, y0=0):
            #         x_rot = (x) * np.cos(theta) + (y) * np.sin(theta) + x0
            #         y_rot = -(x) * np.sin(theta) + (y) * np.cos(theta) + y0
            #         return int(x_rot), int(y_rot)

            #     # Horizontal line (dashed approximation)
            #     num_dashes = 10
            #     for i in range(num_dashes):
            #         start_frac = i / num_dashes
            #         end_frac = (i + 0.5) / num_dashes
            #         x_start = -Lx + 2 * Lx * start_frac
            #         x_end   = -Lx + 2 * Lx * end_frac
            #         y_start, y_end = 0, 0
            #         pt1 = rotate_point(x_start, y_start, theta, x0, y0)
            #         pt2 = rotate_point(x_end, y_end, theta, x0, y0)
            #         cv2.line(display_img, pt1, pt2, (0, 0, 255), 1)

            #     # Vertical line (solid)
            #     pt1 = rotate_point(0, -Ly, theta, x0, y0)
            #     pt2 = rotate_point(0, Ly, theta, x0, y0)
            #     cv2.line(display_img, pt1, pt2, (0, 0, 255), 1)

            #     # Horizontal line (dashed approximation)
            #     num_dashes = 10
            #     for i in range(num_dashes):
            #         start_frac = i / num_dashes
            #         end_frac = (i + 0.5) / num_dashes
            #         x_start = -Lx + 2 * Lx * start_frac
            #         x_end   = -Lx + 2 * Lx * end_frac
            #         y_start, y_end = 0, 0
            #         pt1 = rotate_point(x_start, y_start, theta_fit, x0, y0)
            #         pt2 = rotate_point(x_end, y_end, theta_fit, x0, y0)
            #         cv2.line(display_img, pt1, pt2, (0, 255, 255), 1)

            #     # Vertical line (solid)
            #     pt1 = rotate_point(0, -Ly, theta_fit, x0, y0)
            #     pt2 = rotate_point(0, Ly, theta_fit, x0, y0)
            #     cv2.line(display_img, pt1, pt2, (0, 255, 255), 1)



            #     # Draw text
            #     draw_text(display_img, f"A = {A:.2f} Grayscale Unit", (10, 20), color=text_color)
            #     draw_text(display_img, f"w_x = {w_x_m*1e6:.2f} um", (10, 40), color=text_color)
            #     draw_text(display_img, f"w_y = {w_y_m*1e6:.2f} um", (10, 60), color=text_color)
            #     draw_text(display_img, f"B = {params[5]:.2f} Grayscale Unit", (10, 80), color=text_color)
            #     draw_text(display_img, f"z = {z_position*1000:.2f} mm", (10, 100), color=text_color)
            #     draw_text(display_img, f"Fit theta (Diagnostic only) = {(theta_fit*180/np.pi)%360:.2f} deg", (10, 120), color=text_color)
            #     draw_text(display_img, f"User theta = {theta*180/np.pi:.2f} deg", (10, 140), color=text_color)
            #     draw_text(display_img, f"x0,y0 = {x0} {y0}", (10, 160), color=text_color)



            #     # Let's open a 2nd window to plot the residual 
            #     # roi_data = img #[top_left[1]:bottom_right[1], top_left[0]:bottom_right[0]]
            #     X_fit = np.arange(img.shape[1])
            #     Y_fit = np.arange(img.shape[0])
            #     X, Y = np.meshgrid(X_fit, Y_fit)
            #     gaussian_fit_ROI=gaussian_2d((X,Y),params[0],params[1],params[2],params[3],params[4],params[5],theta).reshape(img.shape)
            #     # print("size roi",np.shape(roi_data),'size gaussian', np.shape(gaussian_fit_ROI))
            #     residual=img-gaussian_fit_ROI


            #     # Normalize residual to 0-255 for color mapping
            #     residual_norm = cv2.normalize(residual, None, 0, 255, cv2.NORM_MINMAX)
            #     residual_uint8 = np.uint8(residual_norm)
            #     residual_color = cv2.applyColorMap(residual_uint8, cv2.COLORMAP_JET)

            #     # create the colorbar
            #     h, w, _ = residual_color.shape
            #     bar_width = 100
            #     scale_factor = 1  # 90% of the height of residual
            #     bar_height = int(h * scale_factor)

            #     gradient = np.linspace(0, 255, bar_height).astype(np.uint8)
            #     gradient = np.tile(gradient[:, None], (1, bar_width))
            #     colorbar = cv2.applyColorMap(gradient, cv2.COLORMAP_JET)

            #     # Pad the colorbar to match residual height
            #     pad_top = (h - bar_height) // 2
            #     pad_bottom = h - bar_height - pad_top
            #     colorbar = cv2.copyMakeBorder(colorbar, pad_top, pad_bottom, 0, 0, cv2.BORDER_CONSTANT, value=[0,0,0])

            #     # printing min and max on the screen
            #     min_val, max_val = np.min(residual), np.max(residual)
            #     # print(h)
            #     cv2.putText(colorbar, f"Max: {max_val:.2f}", (0, 15), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)
            #     cv2.putText(colorbar, f"Min: {min_val:.2f}", (0, h-(int(h/4))), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)

            #     # Concatenate colorbar to the right of the residual
            #     residual_display = cv2.hconcat([residual_color, colorbar])
            #     cv2.imshow("Residual Window", residual_display) # showing a window of the ROI size

            # except Exception as e:
            #     print("Warning: Fit failed for this frame (beam may be absent or too faint). Skipping frame.")
            #     continue  # skip to next frame, no display or processing for this frame
            # cv2.imshow('Beam Profile Fit Window', display_img)
            # key = cv2.waitKey(1) & 0xFF
            
            # cv2.imshow('Beam Profile Fit Window', display_img)
            key = cv2.waitKey(1) & 0xFF

            if key == ord('r'):
                recording = not recording
                print(f"Recording {'ON' if recording else 'OFF'}...")

            if recording:
                wx_temp.append(w_x_m)
                wy_temp.append(w_y_m)
                print(f"Buffered sample: wx={w_x_m*1e6:.2f} um, wy={w_y_m*1e6:.2f} um")

            elif key == ord('R'):
                if wx_temp:
                    w_x_mean = np.mean(wx_temp)
                    w_y_mean = np.mean(wy_temp)
                    w_x_std = max(np.std(wx_temp), 1e-6)
                    w_y_std = max(np.std(wy_temp), 1e-6)
                    wx_list.append(w_x_mean)
                    wy_list.append(w_y_mean)
                    wx_std_list.append(w_x_std)
                    wy_std_list.append(w_y_std)
                    z_list.append(z_position)
                    theta_fit_list.append(theta_fit%360)
                    A_list.append(A)
                    B_list.append(params[5])

                    print(f"Recorded batch: z={z_position:.3f} m, "
                          f"wx={w_x_mean*1e6:.2f}±{w_x_std*1e6:.2f} um, "
                          f"wy={w_y_mean*1e6:.2f}±{w_y_std*1e6:.2f} um, "
                          f"A = {A:.2f} grayscale unit, "
                          f"B = {params[5]:.2f} grayscale unit"
                          f"theta_fit={theta_fit:.2f} deg")

                    plt.clf()
                    plt.errorbar(z_list, np.array(wx_list)*1e6, yerr=np.array(wx_std_list)*1e6, fmt='o', label='wx', capsize=3)
                    plt.errorbar(z_list, np.array(wy_list)*1e6, yerr=np.array(wy_std_list)*1e6, fmt='o', label='wy', capsize=3)
                    plt.xlabel("z position [m]")
                    plt.ylabel("Beam width [um]")
                    plt.legend()
                    plt.title("Live Beam Width vs. z")
                    plt.tight_layout()
                    plt.draw()
                    plt.show(block=False)
                    plt.gcf().canvas.flush_events()
                    wx_temp.clear()
                    wy_temp.clear()

            elif key == ord('n'):
                distance_str = get_distance_input()
                try:
                    dz_inch = float(distance_str)
                    dz_m = dz_inch * 0.0254
                    z_position += dz_m
                    print(f"Moved by {dz_inch:.3f} in = {dz_m:.4f} m, new z = {z_position:.4f} m")
                except Exception:
                    print("Invalid distance.")

            elif key == ord('['):
                theta_val -= theta_step
                print(f"Theta adjusted: {theta_val:.2f} deg")

            elif key == ord(']'):
                theta_val += theta_step
                print(f"Theta adjusted: {theta_val:.2f} deg")

            elif key == ord('f'):
                if len(z_list) < 3:
                    print("Not enough points to fit.")
                    continue

                print("Fitting...")
                z = np.array(z_list)
                wx = np.array(wx_list)
                wy = np.array(wy_list)
                wx_std = np.array(wx_std_list)
                wy_std = np.array(wy_std_list)
                A_val = np.array(A_list)
                B_val = np.array(B_list)

                sol_x, sol_y = fit_beam_profile_ODR(
                    z, wx, z, wy,
                    w0guess=300e-6,
                    z0guess=z_list[0],
                    wx_std=wx_std,
                    wy_std=wy_std,
                    z_std=0.005,
                    title='Beam Profile',
                    print_results=True
                )

                q_x = f"{sol_x[1]:.4e} + i{sol_x[2]:.4e}"
                q_y = f"{sol_y[1]:.4e} + i{sol_y[2]:.4e}"
                print("q-parameter (x):", q_x)
                print("q-parameter (y):", q_y)

                plt.gcf()
                plt.text(0.05, 0.95, f"q_x = {q_x}", transform=plt.gca().transAxes,
                         verticalalignment='top', fontsize=8, color='blue')
                plt.text(0.05, 0.90, f"q_y = {q_y}", transform=plt.gca().transAxes,
                         verticalalignment='top', fontsize=8, color='green')
                plt.draw()

                if ask_to_save():
                    with open(output_file, "w", newline="") as f:
                        writer = csv.writer(f)
                        writer.writerow(["z [m]", "wx [m]", "wy [m]", "wx_std [m]", "wy_std [m]", "theta_fit [deg]", "A [grayscale unit]", "B [grayscale unit]"])
                        for i in range(len(z_list)):
                            writer.writerow([z_list[i], wx_list[i], wy_list[i], wx_std_list[i], wy_std_list[i], theta_fit_list[i], A_list[i], B_list[i]])
                        writer.writerow([])
                        writer.writerow(["q_x", f"'{q_x}'"])
                        writer.writerow(["q_y", f"'{q_y}'"])
                    print(f"Saved to {output_file}")
                break

            elif key == ord('p'):
                try:
                    nodemap = cam.GetNodeMap()

                    # Turn off auto exposure if it’s on
                    exposure_auto = PySpin.CEnumerationPtr(nodemap.GetNode("ExposureAuto"))
                    exposure_auto.SetIntValue(exposure_auto.GetEntryByName("Off").GetValue())

                    # Get the exposure node
                    exposure_time = PySpin.CFloatPtr(nodemap.GetNode("ExposureTime"))

                    if PySpin.IsAvailable(exposure_time) and PySpin.IsWritable(exposure_time):
                        new_exp_str = input("Enter new exposure time in µs: ")
                        try:
                            new_exp = float(new_exp_str)
                            # Clamp to camera limits
                            new_exp = min(max(new_exp, exposure_time.GetMin()), exposure_time.GetMax())
                            exposure_time.SetValue(new_exp)
                            print(f"Exposure manually set to {new_exp:.2f} µs")
                        except ValueError:
                            print("Invalid number, exposure unchanged.")
                    else:
                        print("ExposureTime node not available or not writable.")

                except Exception as e:
                    print(f"Failed to update exposure: {e}")


            elif key == ord('q'):
                print("Quit without fitting.")
                break

            elif key == ord('s'):
                roi_size += roi_size_step
                roi_size = int(roi_size)
                print(f"ROI size adjusted: {roi_size}x{roi_size} pixels")

            elif key ==ord('d'):
                roi_size -= roi_size_step
                roi_size = int(roi_size)
                print(f"ROI size adjusted: {roi_size}x{roi_size} pixels")
            
            elif key == ord('g'):
                gaussian_fit_boolean = not gaussian_fit_boolean
                print(f"Gaussian fit {'ON' if gaussian_fit_boolean else 'OFF'}...")

            if gaussian_fit_boolean:
                try:
                    params = fit_gaussian_roi(img, roi_size=roi_size, downsample=downsample, meshgrid_cache=meshgrid_cache,theta_user=theta_val)
                    x0, y0 = int(params[1]), int(params[2])
                    half = roi_size // 2
                    top_left = (max(0, x0 - half), max(0, y0 - half))
                    bottom_right = (min(img.shape[1], x0 + half), min(img.shape[0], y0 + half))

                    w_x_px = params[3]
                    w_y_px = params[4]

                    w_x_m = w_x_px * pixel_size_m/scaling_factor #1.13 scaling factor by comparing the FLIR camera to the WinCam
                    w_y_m = w_y_px * pixel_size_m/scaling_factor
                    A = params[0]

                    # Use user theta for drawing
                    theta = params[-1]
                    theta_fit=params[-2] 


                    #Adding a grid
                    draw_grid(display_img, step=100, color=(100, 100, 100))  # gray grid
                    # Draw ROI rectangle and ellipse
                    cv2.rectangle(display_img, top_left, bottom_right, rect_color, 1)
                    center = (x0, y0)
                    axes = (int(params[3] * 2), int(params[4] * 2))
                    cv2.ellipse(display_img, center, axes, 0, 0, 360, ellipse_color, 1)

                    # Crosshair parameters
                    # Use ROI size for line lengths
                    Lx = (bottom_right[0] - top_left[0]) // 2  # half width
                    Ly = (bottom_right[1] - top_left[1]) // 2  # half height

                    # Function to rotate points
                    def rotate_point(x, y, theta, x0=0, y0=0):
                        x_rot = (x) * np.cos(theta) + (y) * np.sin(theta) + x0
                        y_rot = -(x) * np.sin(theta) + (y) * np.cos(theta) + y0
                        return int(x_rot), int(y_rot)

                    # Horizontal line (dashed approximation)
                    num_dashes = 10
                    for i in range(num_dashes):
                        start_frac = i / num_dashes
                        end_frac = (i + 0.5) / num_dashes
                        x_start = -Lx + 2 * Lx * start_frac
                        x_end   = -Lx + 2 * Lx * end_frac
                        y_start, y_end = 0, 0
                        pt1 = rotate_point(x_start, y_start, theta, x0, y0)
                        pt2 = rotate_point(x_end, y_end, theta, x0, y0)
                        cv2.line(display_img, pt1, pt2, (0, 0, 255), 1)

                    # Vertical line (solid)
                    pt1 = rotate_point(0, -Ly, theta, x0, y0)
                    pt2 = rotate_point(0, Ly, theta, x0, y0)
                    cv2.line(display_img, pt1, pt2, (0, 0, 255), 1)

                    # Horizontal line (dashed approximation)
                    num_dashes = 10
                    for i in range(num_dashes):
                        start_frac = i / num_dashes
                        end_frac = (i + 0.5) / num_dashes
                        x_start = -Lx + 2 * Lx * start_frac
                        x_end   = -Lx + 2 * Lx * end_frac
                        y_start, y_end = 0, 0
                        pt1 = rotate_point(x_start, y_start, theta_fit, x0, y0)
                        pt2 = rotate_point(x_end, y_end, theta_fit, x0, y0)
                        cv2.line(display_img, pt1, pt2, (0, 255, 255), 1)

                    # Vertical line (solid)
                    pt1 = rotate_point(0, -Ly, theta_fit, x0, y0)
                    pt2 = rotate_point(0, Ly, theta_fit, x0, y0)
                    cv2.line(display_img, pt1, pt2, (0, 255, 255), 1)


                    # Draw text
                    draw_text(display_img, f"A = {A:.2f} Grayscale Unit", (10, 20), color=text_color)
                    draw_text(display_img, f"w_x = {w_x_m*1e6:.2f} um", (10, 40), color=text_color)
                    draw_text(display_img, f"w_y = {w_y_m*1e6:.2f} um", (10, 60), color=text_color)
                    draw_text(display_img, f"B = {params[5]:.2f} Grayscale Unit", (10, 80), color=text_color)
                    draw_text(display_img, f"z = {z_position*1000:.2f} mm", (10, 100), color=text_color)
                    draw_text(display_img, f"Fit theta (Diagnostic only) = {(theta_fit*180/np.pi)%360:.2f} deg", (10, 120), color=text_color)
                    draw_text(display_img, f"User theta = {theta*180/np.pi:.2f} deg", (10, 140), color=text_color)
                    draw_text(display_img, f"x0,y0 = {x0} {y0}", (10, 160), color=text_color)


                    # Let's open a 2nd window to plot the residual 
                    # roi_data = img #[top_left[1]:bottom_right[1], top_left[0]:bottom_right[0]]
                    X_fit = np.arange(img.shape[1])
                    Y_fit = np.arange(img.shape[0])
                    X, Y = np.meshgrid(X_fit, Y_fit)
                    gaussian_fit_ROI=gaussian_2d((X,Y),params[0],params[1],params[2],params[3],params[4],params[5],theta).reshape(img.shape)
                    # print("size roi",np.shape(roi_data),'size gaussian', np.shape(gaussian_fit_ROI))
                    residual=img-gaussian_fit_ROI


                    # Normalize residual to 0-255 for color mapping
                    residual_norm = cv2.normalize(residual, None, 0, 255, cv2.NORM_MINMAX)
                    residual_uint8 = np.uint8(residual_norm)
                    residual_color = cv2.applyColorMap(residual_uint8, cv2.COLORMAP_JET)

                    # create the colorbar
                    h, w, _ = residual_color.shape
                    bar_width = 100
                    scale_factor = 1  # 90% of the height of residual
                    bar_height = int(h * scale_factor)

                    gradient = np.linspace(0, 255, bar_height).astype(np.uint8)
                    gradient = np.tile(gradient[:, None], (1, bar_width))
                    colorbar = cv2.applyColorMap(gradient, cv2.COLORMAP_JET)

                    # Pad the colorbar to match residual height
                    pad_top = (h - bar_height) // 2
                    pad_bottom = h - bar_height - pad_top
                    colorbar = cv2.copyMakeBorder(colorbar, pad_top, pad_bottom, 0, 0, cv2.BORDER_CONSTANT, value=[0,0,0])

                    # printing min and max on the screen
                    min_val, max_val = np.min(residual), np.max(residual)
                    # print(h)
                    cv2.putText(colorbar, f"Max: {max_val:.2f}", (0, 15), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)
                    cv2.putText(colorbar, f"Min: {min_val:.2f}", (0, h-(int(h/4))), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)

                    # Concatenate colorbar to the right of the residual
                    residual_display = cv2.hconcat([residual_color, colorbar])
                    cv2.imshow("Residual Window", residual_display) # showing a window of the ROI size
                    cv2.imshow('Beam Profile Fit Window', display_img)

                except Exception as e:
                    print("Warning: Fit failed for this frame (beam may be absent or too faint). Skipping frame.")
                    continue  # skip to next frame, no display or processing for this frame
                
            elif gaussian_fit_boolean == False:
                cv2.imshow('Beam Profile Fit Window', display_img)

        

    finally:
        cam.EndAcquisition()
        cam.DeInit()
        del cam
        cam_list.Clear()
        system.ReleaseInstance()
        cv2.destroyAllWindows()
        plt.ioff()
        plt.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Live Gaussian beam profiling with camera.\n\n"
                                     "Example:\n"
                                     "  python beam_profile.py --roi-size 400 --downsample 2 --exposure auto --gain auto --pixel-size 6.9 --output my_beam_scan.csv",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--roi-size', type=int, default=300, help='ROI size in pixels')
    parser.add_argument('--downsample', type=int, default=2, help='Downsampling factor')
    parser.add_argument('--exposure', default='auto', help='Camera exposure (µs)')
    parser.add_argument('--gain', default='auto', help='Camera gain (dB)')
    parser.add_argument('--pixel-size', type=float, default=6.9, help='Pixel size in um (default 6.9)')
    parser.add_argument('--output', default="beam_profile.csv", help='Output CSV filename')
    parser.add_argument('--theta_val', type=float, default=0, help='angle to fit in deg')
    parser.add_argument('--scaling_factor', type=float, default=1, help='Scaling factor')
    parser.add_argument('--mode', choices=['gray', 'heatmap'], default='gray', help='Display mode for live camera (default: gray)')
    args = parser.parse_args()

    beam_profile_fit(
        roi_size=args.roi_size,
        downsample=args.downsample,
        exposure=args.exposure,
        gain=args.gain,
        pixel_size_um=args.pixel_size,
        output_file=args.output,
        theta_val=args.theta_val,
        scaling_factor=args.scaling_factor,
        mode=args.mode
    )